@@ -37,6 +37,7 @@
 #include <pango/pango.h>
 #include <cairo/cairo.h>
 
+#include <gtk/gtksocket.h>
 #include <gtk/gtkwindow.h>
 #include <gtk/gtkcomboboxentry.h>
 #include <gtk/gtktoolbar.h>
@@ -72,6 +73,7 @@ static HildonIMTrigger trigger = 0;
 static HildonIMCommitMode commit_mode = 0;
 static gboolean is_internal_widget = FALSE;
 static gboolean internal_reset = FALSE;
+static gboolean enter_on_focus_pending = FALSE;
 
 typedef enum {
   HILDON_IM_SHIFT_STICKY_MASK     = 1 << 0,
@@ -850,6 +852,9 @@ client_message_filter(GdkXEvent *xevent,GdkEvent *event,
           hildon_im_context_send_fake_key(GDK_KP_Enter, TRUE);
           hildon_im_context_send_fake_key(GDK_KP_Enter, FALSE);
           break;
+        case HILDON_IM_CONTEXT_ENTER_ON_FOCUS:
+          enter_on_focus_pending = TRUE;
+          break;
         case HILDON_IM_CONTEXT_CONFIRM_SENTENCE_START:
           hildon_im_context_check_sentence_start(self);
           break;
@@ -1180,6 +1185,13 @@ hildon_im_context_focus_in(GtkIMContext *context)
   }
 
   hildon_im_context_send_command(self, HILDON_IM_SETCLIENT);
+
+  if (enter_on_focus_pending)
+  {
+    hildon_im_context_send_fake_key(GDK_KP_Enter, TRUE);
+    hildon_im_context_send_fake_key(GDK_KP_Enter, FALSE);
+    enter_on_focus_pending = FALSE;
+  }
 }
 
 static void
@@ -1656,6 +1668,7 @@ hildon_im_context_filter_keypress(GtkIMContext *context, GdkEventKey *event)
                                      gdk_unicode_to_keyval(c),
                                      event->hardware_keycode);
     self->last_internal_change = TRUE;
+    self->auto_upper = FALSE;
     g_signal_emit_by_name (self, "commit", utf8);
 
     return TRUE;
@@ -2068,7 +2081,26 @@ hildon_im_context_send_command(HildonIMContext *self,
   if (cmd != HILDON_IM_HIDE)
   {
     msg->input_window = GDK_WINDOW_XID(input_window);
-    msg->app_window = GDK_WINDOW_XID(gdk_window_get_toplevel(input_window));
+
+    /* When the client widget is a child of GtkPlug, the application can
+       override the ID of the window the IM will be set transient to */
+    if (GTK_IS_PLUG(gtk_widget_get_toplevel(self->client_gtk_widget)))
+    {
+      GtkPlug *plug = (GtkPlug *)gtk_widget_get_toplevel(self->client_gtk_widget);
+      guint32 transient_window_xid;
+
+      transient_window_xid =
+        GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(plug),
+                                           "hildon_im_transient_xid"));
+      if (transient_window_xid)
+        msg->app_window = transient_window_xid;
+      else
+        msg->app_window = GDK_WINDOW_XID(gdk_window_get_toplevel(input_window));
+    }
+    else
+    {
+      msg->app_window = GDK_WINDOW_XID(gdk_window_get_toplevel(input_window));
+    }
   }
 
   msg->cmd = cmd;
